class ContextManager:
    def __init__(self, result=None):
        self.context_depth = 0
        self.result = result

    def create_context(self, for_target):
        name = getattr(for_target, "__name__", for_target)
        return Context(name, self)

    def enter(self, context):
        self.context_depth += 1

        enterContext = getattr(self.result, "enterContext", None)
        if enterContext is not None:
            enterContext(context, depth=self.context_depth)

    def exit(self):
        self.context_depth -= 1

        exitContext = getattr(self.result, "exitContext", None)
        if exitContext is not None:
            exitContext(depth=self.context_depth)


class Context:
    def __init__(self, name, manager):
        self.manager = manager
        self.name = name

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.name == other.name

    def __ne__(self, other):
        return not self == other

    def __enter__(self):
        """
        Enter the context.

        """

        self.manager.enter(self)

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Exit the context.

        """

        self.manager.exit()



from ivoire import describe
from ivoire.manager import Context
from ivoire.spec.util import ExampleWithPatch, mock

with describe(Context, Example=ExampleWithPatch) as it:

    @it.before
    def before(test):
        test.manager = mock.Mock()
        test.context = Context("a test context", test.manager)

    with it("calls its manager when used as a context manager") as test:
        with test.context:
            test.manager.enter.assert_called_once_with(test.context)
        test.manager.exit.assert_called_once_with()



from ivoire import ContextManager, context, describe
from ivoire.manager import Context
from ivoire.spec.util import ExampleWithPatch, mock

with describe(ContextManager, Example=ExampleWithPatch) as it:

    @it.before
    def before(test):
        test.result = mock.Mock()
        test.manager = ContextManager(test.result)
        test.context = test.manager.create_context("a test context")

    with context(context):
        with it("creates contexts") as test:
            context = test.manager.create_context("a test context")
            test.assertEqual(context, Context("a test context", test.manager))

        with it("is a bit nasty and tries to get __name__s") as test:

            def foo():
                pass

            context = test.manager.create_context(foo)
            test.assertEqual(context, Context("foo", test.manager))

    with it("starts off at a global context depth of 0") as test:
        test.assertEqual(test.manager.context_depth, 0)

    with it("enters and exits contexts") as test:
        test.manager.enter(test.context)
        test.result.enterContext.assert_called_once_with(test.context, depth=1)

        test.manager.exit()
        test.result.exitContext.assert_called_once_with(depth=0)

    with it("doesn't call methods if the result doesn't know how") as test:
        del test.result.enterContext, test.result.exitContext

        test.manager.enter(test.context)
        test.manager.exit()
        test.assertFalse(test.result.method_calls)



"""
Test the standalone running of Ivoire examples.

A **warning** about testing in this module: Testing Ivoire Examples means
testing a thing that is swallowing (and recording) *all* exceptions. Make sure
all of your assertions are outside of example blocks so that they are handled
by the surrounding test case.

"""


from unittest import TestCase
import sys

from ivoire.standalone import describe
from ivoire.tests.util import PatchMixin, mock


class TestDescribeTests(TestCase, PatchMixin):
    def setUp(self):
        self.result = self.patch("ivoire.current_result", shouldStop=False)
        self.describes = describe
        self.it = describe(self.describes)

    def test_it_adds_an_example(self):
        with self.it as it:
            with it("does a thing") as test:
                pass
        self.assertEqual(it.examples, [test])

    def test_it_can_pass(self):
        with self.it as it:
            with it("does a thing") as test:
                pass

        self.result.assert_has_calls(
            [
                mock.call.enterGroup(self.it),
                mock.call.startTest(test),
                mock.call.addSuccess(test),
                mock.call.stopTest(test),
                mock.call.exitGroup(self.it),
            ],
        )

    def test_it_can_fail(self):
        with self.it as it:
            with it("does a thing") as test:
                try:
                    test.fail()
                except Exception:
                    exc_info = sys.exc_info()
                    raise

        self.result.assert_has_calls(
            [
                mock.call.enterGroup(self.it),
                mock.call.startTest(test),
                mock.call.addFailure(test, exc_info),
                mock.call.stopTest(test),
                mock.call.exitGroup(self.it),
            ],
        )

    def test_it_can_error(self):
        with self.it as it:
            with it("does a thing") as test:
                try:
                    raise IndexError
                except IndexError:
                    exc_info = sys.exc_info()
                    raise

        self.result.assert_has_calls(
            [
                mock.call.enterGroup(self.it),
                mock.call.startTest(test),
                mock.call.addError(test, exc_info),
                mock.call.stopTest(test),
                mock.call.exitGroup(self.it),
            ],
        )

    def test_it_does_not_swallow_KeyboardInterrupts(self):
        with self.assertRaises(KeyboardInterrupt):
            with self.it as it:
                with it("does a thing"):
                    raise KeyboardInterrupt

    def test_it_exits_the_group_if_begin_errors(self):
        self.ran_test = False

        with self.it as it:

            @it.before
            def before(test):
                raise RuntimeError("Buggy before.")

            example = it("should not be run")
            with example:
                self.ran_test = True

        self.assertEqual(
            self.result.mock_calls,
            [
                mock.call.enterGroup(self.it),
                mock.call.startTest(example),
                mock.call.addError(example, mock.ANY),  # traceback object
                mock.call.stopTest(example),
                mock.call.exitGroup(self.it),
            ],
        )
        self.assertFalse(self.ran_test)

    def test_it_runs_befores(self):
        with self.it as it:

            @it.before
            def before(test):
                test.foo = 12

            with it("should have set foo") as test:
                foo = test.foo
        self.assertEqual(foo, 12)

    # XXX: There's a few more cases here that should be tested

    def test_it_runs_afters(self):
        self.foo = None

        with self.it as it:

            @it.after
            def after(test):
                self.foo = 12

            with it("should have set foo after"):
                foo = self.foo

        self.assertEqual(foo, None)
        self.assertEqual(self.foo, 12)

    def test_it_runs_cleanups(self):
        with self.it as it:
            with it("does a thing") as test:
                doCleanups = self.patchObject(test, "doCleanups")
        self.assertTrue(doCleanups.called)

    def test_it_respects_shouldStop(self):
        with self.it as it:
            with it("does a thing"):
                self.result.shouldStop = True
            self.fail("should have stopped already!")  # pragma: no cover

    def test_it_can_skip(self):
        with self.it as it:
            with it("should skip this test") as test:
                test.skip_if(False, reason="A bad one")
                test.skip_if(True, reason="A good one")
                test.fail("Should have skipped!")

        self.assertEqual(
            self.result.method_calls,
            [
                mock.call.enterGroup(self.it),
                mock.call.startTest(test),
                mock.call.addSkip(test, "A good one"),
                mock.call.stopTest(test),
                mock.call.exitGroup(self.it),
            ],
        )

    def test_it_only_calls_enterGroup_if_result_knows_how(self):
        del self.result.enterGroup

        with self.it:
            pass

        self.assertEqual(
            self.result.method_calls,
            [mock.call.exitGroup(self.it)],
        )

    def test_it_only_calls_exitGroup_if_result_knows_how(self):
        del self.result.exitGroup

        with self.it:
            pass

        self.assertEqual(
            self.result.method_calls,
            [mock.call.enterGroup(self.it)],
        )



from pathlib import Path

import nox

ROOT = Path(__file__).parent
DOCS = ROOT / "docs"
IVOIRE = ROOT / "ivoire"


nox.options.sessions = []


def session(default=True, **kwargs):
    def _session(fn):
        if default:
            nox.options.sessions.append(kwargs.get("name", fn.__name__))
        return nox.session(**kwargs)(fn)

    return _session


@session(python=["3.8", "3.9", "3.10", "3.11", "pypy3"])
def tests(session):
    session.install("virtue", str(ROOT))
    session.run("virtue", "ivoire")
    session.run("ivoire", IVOIRE)


@session()
def audit(session):
    session.install("pip-audit", str(ROOT))
    session.run("python", "-m", "pip_audit")


@session(tags=["build"])
def build(session):
    session.install("build")
    tmpdir = session.create_tmp()
    session.run("python", "-m", "build", str(ROOT), "--outdir", tmpdir)


@session(tags=["style"])
def readme(session):
    session.install("build", "twine")
    tmpdir = session.create_tmp()
    session.run("python", "-m", "build", str(ROOT), "--outdir", tmpdir)
    session.run("python", "-m", "twine", "check", tmpdir + "/*")


@session(tags=["style"])
def style(session):
    session.install(
        "flake8",
        "flake8-broken-line",
        "flake8-bugbear",
        "flake8-commas",
        "flake8-quotes",
        "flake8-tidy-imports",
    )
    session.run(
        "python",
        "-m",
        "flake8",
        str(IVOIRE),
        __file__,
    )


@session()
def typing(session):
    session.install("mypy", str(ROOT))
    session.run("python", "-m", "mypy", str(IVOIRE))
